using System.IdentityModel.Tokens.Jwt;
using System.Security.Cryptography;
using System.Text;
using FitConnect.Models.v1.Api;
using IdentityModel;
using Jose;
using Microsoft.Extensions.Logging;
using Microsoft.IdentityModel.JsonWebTokens;
using Microsoft.IdentityModel.Tokens;
using Newtonsoft.Json;

namespace FitConnect.Encryption;

public class KeySet {
    public string? PrivateKeyDecryption { get; set; }
    public string? PrivateKeySigning { get; set; }
    public string? PublicKeyEncryption { get; set; }
    public string? PublicKeySignatureVerification { get; set; }
}

public class FitEncryption {
    private readonly IEncryptor _encryptor = new JoseEncryptor();
    private readonly ILogger? _logger;
    private JwtHeader _jwtHeader;

    internal FitEncryption(ILogger? logger) {
        _logger = logger;
    }

    public FitEncryption(string? privateKeyDecryption, string? privateKeySigning,
        string? publicKeyEncryption, string? publicKeySignatureVerification, ILogger? logger) :
        this(logger) {
        PrivateKeyDecryption = privateKeyDecryption;
        PrivateKeySigning = privateKeySigning;
        PublicKeyEncryption = publicKeyEncryption;
        PublicKeySignatureVerification = publicKeySignatureVerification;
    }

    public FitEncryption(
        KeySet keySet,
        ILogger? logger) : this(logger) {
        PrivateKeyDecryption = keySet.PrivateKeyDecryption;
        PrivateKeySigning = keySet.PrivateKeySigning;
        PublicKeyEncryption = keySet.PublicKeyEncryption;
        PublicKeySignatureVerification = keySet.PublicKeySignatureVerification;
    }

    public string? PrivateKeyDecryption { get; set; }
    public string? PrivateKeySigning { get; set; }
    public string? PublicKeyEncryption { get; set; }
    public string? PublicKeySignatureVerification { get; set; }

    public (string plainText, byte[] plainBytes, byte[] tag)
        Decrypt(string cypherText, string key) {
        return _encryptor.Decrypt(key, cypherText);
    }

    public (string plainText, byte[] plainBytes, byte[] tag) Decrypt(string cypherText) {
        if (PrivateKeyDecryption == null)
            throw new InvalidOperationException("PrivateKey is not provided");

        return Decrypt(cypherText, PrivateKeyDecryption);
    }


    public string Encrypt(string plainText, string key) {
        return _encryptor.Encrypt(key, plainText);
    }

    public string Encrypt(string plain) {
        if (PublicKeyEncryption == null)
            throw new InvalidOperationException("PublicKey is not provided");

        return Encrypt(plain, PublicKeyEncryption);
    }

    public string Encrypt(byte[] plain, string key) {
        return _encryptor.Encrypt(key, plain);
    }

    public string Encrypt(byte[] plain) {
        if (PublicKeyEncryption == null)
            throw new InvalidOperationException("PublicKey is not provided");

        return Encrypt(plain, PublicKeyEncryption);
    }

    public string CreateRejectSecurityEventToken(string submissionId,
        string caseId,
        string destinationId, Problems[]? problemsArray) =>
        CreateSecurityEventToken(submissionId, caseId, destinationId,
            "https://schema.fitko.de/fit-connect/events/reject-submission",
            new { problems = problemsArray });

    public string CreateAcceptSecurityEventToken(string submissionSubmissionId,
        string submissionCaseId, string submissionDestinationId) {
        var payload = new {
            authenticationTags = new {
                data = "", metadata = "",
                attachments = new Dictionary<string, string>() { { "", "" } }
            }
        };
        
        return CreateSecurityEventToken(submissionSubmissionId, submissionCaseId,
            submissionDestinationId,
            "https://schema.fitko.de/fit-connect/events/accept-submission", payload);
    }


    private string CreateSecurityEventToken(string submissionId,
        string caseId,
        string destinationId,
        string eventName, object? content) {
        var signingKey = new JsonWebKey(PrivateKeySigning);
        var transactionId = "case:" + caseId;
        var subject = "submission:" + submissionId;

        _jwtHeader =
            new JwtHeader(new SigningCredentials(signingKey, "PS512")) {
                //{ "typ", "secevent+jwt" },
                //  { "kid", signingKey.KeyId },
                //  { "alg", "PS512" }
            };


        _jwtHeader["typ"] = "secevent+jwt";

        var token = new JwtSecurityToken(
            _jwtHeader, new JwtPayload {
                { "sub", subject },
                // { "$schema", "https://schema.fitko.de/fit-connect/set-payload/1.0.0/set-payload.schema.json" },
                { "jti", Guid.NewGuid().ToString() },
                { "iat", DateTime.UtcNow.ToEpochTime() },
                { "iss", destinationId }, {
                    "events",
                    new Dictionary<string, object> {
                        { eventName, content }
                    }
                },
                { "txn", transactionId }
            }) {
            SigningKey = signingKey
        };

        var handler = new JwtSecurityTokenHandler();
        return handler.WriteToken(token);
    }

    public static string CalculateHash(string data) {
        return ByteToHexString(SHA512.Create().ComputeHash(Encoding.UTF8.GetBytes(data)));
    }

    private static string ByteToHexString(IEnumerable<byte> data) {
        var sb = new StringBuilder();
        foreach (var b in data) sb.Append(b.ToString("x2"));

        return sb.ToString();
    }


    public static bool VerifyJwt(string signature, JsonWebKeySet keySet, ILogger? logger = null) {
        foreach (var key in keySet.Keys) {
            try {
                var valid = VerifyJwt(signature, key, logger);
                if (valid) return true;
            }
            catch (Exception e) {
                logger?.LogError(e, "Error verifying JWT");
            }
        }

        return false;
    }

    public static bool VerifyJwt(string signature, string secret,
        ILogger? logger = null) {
        return VerifyJwt(signature,
            new JsonWebKey(secret), logger);
    }

    public static bool VerifyJwt(string signature, JsonWebKey key,
        ILogger? logger = null) {
        var tokenHandler = new JsonWebTokenHandler();

        logger?.LogTrace("Verifying JWT {Token}", signature);

        // if (key.KeySize != 4096)
        //     throw new Exception("Key size must be 4096");

        var result = tokenHandler.ValidateToken(signature, new TokenValidationParameters {
            ValidAlgorithms = new[] { "PS512" },
            IssuerSigningKey = key,
            ValidateIssuerSigningKey = true,
            ValidateIssuer = false,
            ValidateAudience = false,
            ValidateLifetime = true,
            ValidateActor = true,
            ValidateSignatureLast = true,
            ValidateTokenReplay = true,
            ValidateWithLKG = true,
            RequireAudience = true,
            RefreshBeforeValidation = true,
            RequireSignedTokens = true,
            RequireExpirationTime = false,
            IgnoreTrailingSlashWhenValidatingAudience = true,
            TryAllIssuerSigningKeys = true

            // set clockskew to zero so tokens expire exactly at token expiration time (instead of 5 minutes later)
            // ClockSkew = TimeSpan.Zero
        });

        var error = result.Exception?.Message;

        if (error != null)
            logger?.LogWarning("Validation error {Error}", error);
        else
            logger?.LogInformation("JWT is valid");

        return result.IsValid;
    }
}
